/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/crc.h                                                  *
 * Created:       2000-09-26 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2000-2007 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: crc.h 71 2007-05-04 20:50:26Z hampa $ */


#ifndef LECO_CRC_H
#define LECO_CRC_H 1


#define CRC_CRC16 0x8003U
#define CRC_CCITT 0x1021U
#define CRC_CRC32 0x04c11db7UL


/*!***************************************************************************
 * @short The CRC-16 context
 *****************************************************************************/
typedef struct {
	unsigned short reg;
	unsigned short gen;
	unsigned long  cnt[2];
	unsigned short tab[256];
} crc16_t;


/*!***************************************************************************
 * @short The CRC-32 context
 *****************************************************************************/
typedef struct {
	unsigned long reg;
	unsigned long gen;
	unsigned long cnt[2];
	unsigned long tab[256];
} crc32_t;


/*!***************************************************************************
 * @short Initialize a CRC-16 context
 * @param crc The CRC-16 context
 * @param gen The generator polynomial
 *
 * Initializes the context and creates the crc table.
 *****************************************************************************/
void crc16_init (crc16_t *crc, unsigned short gen);

void crc16_reset (crc16_t *crc);

/*!***************************************************************************
 * @short Copy a CRC-16 context
 *****************************************************************************/
void crc16_copy (crc16_t *dst, const crc16_t *src);

/*!***************************************************************************
 * @short  Get the CRC register
 * @return The CRC16 checksum
 *
 * crc16_done() must be called before this.
 *****************************************************************************/
unsigned short crc16_get_reg (const crc16_t *crc);

/*!***************************************************************************
 * @short Calculate a CRC-16 checksum for a byte buffer.
 * @param crc The CRC-16 context
 * @param buf The buffer
 * @param cnt The buffer size
 *
 * This function can be called multiple times.
 *****************************************************************************/
void crc16_calc (crc16_t *crc, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short Finish calculating a CRC-16 checksum
 * @param crc   The CRC-16 context
 * @param posix If true then use the posix algorithm
 *
 * This function should be called after the last call to crc16_calc() and
 * before crc16_get_reg().
 *****************************************************************************/
void crc16_done (crc16_t *crc, int posix);


/*!***************************************************************************
 * @short Initialize a CRC-32 context
 *****************************************************************************/
void crc32_init (crc32_t *crc, unsigned long gen);

void crc32_reset (crc32_t *crc);

/*!***************************************************************************
 * @short Copy a CRC-32 context
 *****************************************************************************/
void crc32_copy (crc32_t *dst, const crc32_t *src);

/*!***************************************************************************
 * @short Get the CRC register.
 *****************************************************************************/
unsigned long crc32_get_reg (const crc32_t *crc);

/*!***************************************************************************
 * @short Calculate a CRC-32 checksum for a byte buffer.
 * @param crc The CRC-32 context
 * @param buf The buffer
 * @param cnt The buffer size
 *
 * This function can be called multiple times.
 *****************************************************************************/
void crc32_calc (crc32_t *crc, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short Finish calculating a CRC-32 checksum
 * @param crc   The CRC-32 context
 * @param posix If true then use the posix algorithm
 *
 * This function should be called after the last call to crc32_calc() and
 * before crc32_get_reg().
 *****************************************************************************/
void crc32_done (crc32_t *crc, int posix);


#endif
