/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecofile.c                                             *
 * Created:       2005-08-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecofile.c 65 2006-12-01 03:11:36Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "block.h"
#include "crc.h"
#include "lecofile.h"


int lec_get_pos (FILE *fp, uint64_t *pos)
{
	off_t off;

	off = ftello (fp);
	if (off == (off_t) -1) {
		return (1);
	}

	*pos = (uint64_t) off;

	return (0);
}

int lec_set_pos (FILE *fp, uint64_t pos)
{
	if (fseeko (fp, pos, SEEK_SET)) {
		return (1);
	}

	return (0);
}

int lec_get_fsize (FILE *fp, uint64_t *size)
{
	if (fseeko (fp, 0, SEEK_END)) {
		return (1);
	}

	if (lec_get_pos (fp, size)) {
		return (1);
	}

	return (0);
}

unsigned long lec_read (FILE *fp, uint64_t pos, void *buf, unsigned long cnt)
{
	if (lec_set_pos (fp, pos)) {
		return (1);
	}

	return (fread (buf, 1, cnt, fp));
}

unsigned long lec_write (FILE *fp, uint64_t pos, const void *buf, unsigned long cnt)
{
	if (lec_set_pos (fp, pos)) {
		return (1);
	}

	return (fwrite (buf, 1, cnt, fp));
}

unsigned char *lec_read_buf (FILE *fp, uint64_t pos, unsigned long cnt)
{
	unsigned char *buf;

	buf = malloc (cnt);
	if (buf == NULL) {
		return (NULL);
	}

	if (lec_read (fp, pos, buf, cnt) != cnt) {
		free (buf);
		return (NULL);
	}

	return (buf);
}

uint16_t lec_get_uint16 (const void *buf, unsigned i)
{
	uint16_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];

	return (val);
}

uint32_t lec_get_uint32 (const void *buf, unsigned i)
{
	uint32_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];
	val = (val << 8) | tmp[2];
	val = (val << 8) | tmp[3];

	return (val);
}

uint64_t lec_get_uint64 (const void *buf, unsigned i)
{
	uint64_t            val;
	const unsigned char *tmp = (const unsigned char *) buf + i;

	val = tmp[0];
	val = (val << 8) | tmp[1];
	val = (val << 8) | tmp[2];
	val = (val << 8) | tmp[3];
	val = (val << 8) | tmp[4];
	val = (val << 8) | tmp[5];
	val = (val << 8) | tmp[6];
	val = (val << 8) | tmp[7];

	return (val);
}

void lec_set_uint16 (void *buf, unsigned i, uint16_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 8) & 0xff;
	tmp[1] = val & 0xff;
}

void lec_set_uint32 (void *buf, unsigned i, uint32_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >> 8) & 0xff;
	tmp[3] = val & 0xff;
}

void lec_set_uint64 (void *buf, unsigned i, uint64_t val)
{
	unsigned char *tmp = (unsigned char *) buf + i;

	tmp[0] = (val >> 56) & 0xff;
	tmp[1] = (val >> 48) & 0xff;
	tmp[2] = (val >> 40) & 0xff;
	tmp[3] = (val >> 32) & 0xff;
	tmp[4] = (val >> 24) & 0xff;
	tmp[5] = (val >> 16) & 0xff;
	tmp[6] = (val >> 8) & 0xff;
	tmp[7] = val & 0xff;
}

uint32_t lec_get_crc (const void *buf, unsigned long cnt)
{
	static int     ok = 0;
	static crc32_t crc;

	if (ok == 0) {
		crc32_init (&crc, CRC_CRC32);
		ok = 1;
	}

	crc32_reset (&crc);
	crc32_calc (&crc, buf, cnt);
	crc32_done (&crc, 1);

	return (crc32_get_reg (&crc));
}

void lecf_init (lec_file_t *lf, FILE *fp)
{
	lf->fp = fp;
	lf->base = 0;

	lf->hofs = 0;

	lf->magic = LECO_MAGIC;
	lf->version = 0;
	lf->hsize = 80;
	lf->bsize = 0;
	lf->size = 0;
	lf->wsize = 16;
	lf->dofs = 80;
	lf->eofs = 0;
	lf->rofs = 0;
	lf->bcnt = 0;
	lf->ecnt = 0;
	lf->zofs = 0;
	lf->fsize = 0;
	lf->fcrc = 0;
	lf->dcrc = 0;
	lf->hcrc = 0;

	memset (lf->header, 0, LECO_HEADER);
	memset (lf->md5, 0, 16);
}

static
int lecf_set_base_end (lec_file_t *lf)
{
	uint64_t      ofs1, ofs2;
	unsigned char buf[16];

	if (lec_get_fsize (lf->fp, &ofs1)) {
		return (1);
	}

	if (ofs1 < 16) {
		return (1);
	}

	ofs1 -= 16;

	if (lec_read (lf->fp, ofs1, buf, 16) != 16) {
		return (1);
	}

	if (memcmp (buf, LECO_MAGIC4, 8) != 0) {
		return (1);
	}

	ofs2 = lec_get_uint64 (buf, 8);

	if (ofs1 < ofs2) {
		return (1);
	}

	lf->base = ofs1 - ofs2;

	return (0);
}

static
int lecf_open_ofs (lec_file_t *lf, uint64_t hofs)
{
	unsigned long cnt;
	unsigned char *buf;

	buf = &lf->header[0];

	if (lecf_read (lf, hofs, buf, 8) != 8) {
		return (1);
	}

	lf->magic = lec_get_uint32 (buf, 0);
	if (lf->magic != LECO_MAGIC) {
		return (1);
	}

	lf->version = lec_get_uint16 (buf, 4);
	if (lf->version != 0) {
		return (1);
	}

	lf->hsize = lec_get_uint16 (buf, 6);
	if (lf->hsize < 64) {
		return (1);
	}

	cnt = (lf->hsize < LECO_HEADER) ? (lf->hsize - 8) : (LECO_HEADER - 8);

	if (lecf_read (lf, hofs + 8, buf + 8, cnt) != cnt) {
		return (1);
	}

	lf->bsize = lec_get_uint32 (buf, 8);
	lf->wsize = lec_get_uint32 (buf, 12);
	lf->dofs = lec_get_uint64 (buf, 16);
	lf->eofs = lec_get_uint64 (buf, 24);
	lf->bcnt = lec_get_uint32 (buf, 32);
	lf->ecnt = lec_get_uint32 (buf, 36);
	lf->fsize = lec_get_uint64 (buf, 40);

	lf->size = ((lf->wsize + 7) / 8) * lf->bsize;

	lf->rofs = ((lf->eofs + lf->ecnt * (lf->size + 8UL) + 15) & ~15ULL) + 16;

	memcpy (lf->md5, buf + 48, 16);

	if (lf->hsize >= 80) {
		lf->fcrc = lec_get_uint32 (buf, 68);
		lf->dcrc = lec_get_uint32 (buf, 72);
		lf->hcrc = lec_get_uint32 (buf, 76);

		lec_set_uint32 (buf, 76, 0);

		if (lec_get_crc (buf, lf->hsize) != lf->hcrc) {
			return (1);
		}

		lec_set_uint32 (buf, 76, lf->hcrc);
	}
	else {
		lf->fcrc = 0;
		lf->dcrc = 0;
		lf->hcrc = 0;
	}

	lf->hofs = hofs;

	return (0);
}

static
int lecf_open_scan (lec_file_t *lf)
{
	uint64_t ofs;

	if (lecf_scan (lf, LECO_MAGIC3, 8, &ofs)) {
		return (1);
	}

	if (lecf_open_ofs (lf, ofs + 16) == 0) {
		lf->rofs = ofs + 16;
		return (0);
	}

	return (1);
}

int lecf_open (lec_file_t *lf, FILE *fp)
{
	lecf_init (lf, fp);

	if (lecf_open_ofs (lf, 0) == 0) {
		return (0);
	}

	if (lecf_set_base_end (lf) == 0) {
		if (lecf_open_ofs (lf, 0) == 0) {
			return (0);
		}

		if (lecf_open_scan (lf) == 0) {
			return (0);
		}

		lf->base = 0;
	}

	if (lecf_open_scan (lf) == 0) {
		return (0);
	}

	return (1);
}

unsigned long lecf_read (lec_file_t *lf, uint64_t pos, void *buf, unsigned long cnt)
{
	return (lec_read (lf->fp, lf->base + pos, buf, cnt));
}

unsigned long lecf_write (lec_file_t *lf, uint64_t pos, const void *buf, unsigned long cnt)
{
	return (lec_write (lf->fp, lf->base + pos, buf, cnt));
}

unsigned char *lecf_read_buf (lec_file_t *lf, uint64_t pos, unsigned long cnt)
{
	return (lec_read_buf (lf->fp, lf->base + pos, cnt));
}

int lecf_scan (lec_file_t *lf, const void *src, unsigned cnt, uint64_t *ofs)
{
	unsigned            i;
	unsigned            bidx, bcnt;
	const unsigned char *tmp;
	uint64_t            bofs;
	unsigned char       buf[256];

	if (cnt > 16) {
		return (1);
	}

	tmp = (const unsigned char *) src;

	bcnt = 0;
	bofs = 0;

	while (1) {
		bcnt += lecf_read (lf, bofs + bcnt, buf + bcnt, 256 - bcnt);
		if (bcnt < cnt) {
			return (1);
		}

		bidx = 0;
		while (bidx <= (bcnt - cnt)) {
			for (i = 0; i < cnt; i++) {
				if (buf[bidx + i] != tmp[i]) {
					break;
				}
			}

			if (i >= cnt) {
				*ofs = bofs + bidx;
				return (0);
			}

			bidx += 1;
		}

		for (i = bidx; i < bcnt; i++) {
			buf[i - bidx] = buf[i];
		}

		bcnt -= bidx;
		bofs += bidx;
	}

	return (1);
}

void lecf_calc_dofs (lec_file_t *lf)
{
	lf->dofs = ((LECO_HEADER + 15) & ~15ULL) + 16;
}

void lecf_calc_eofs (lec_file_t *lf)
{
	lf->eofs = ((lf->dofs + 4UL * lf->bcnt + 15) & ~15ULL) + 16;
}

void lecf_calc_rofs (lec_file_t *lf)
{
	lf->rofs = ((lf->eofs + lf->ecnt * (lf->size + 8UL) + 15) & ~15ULL) + 16;
	lf->zofs = lf->rofs;
}

int lecf_write_header (lec_file_t *lf)
{
	unsigned long crc;
	unsigned char *buf;

	buf = &lf->header[0];

	lec_set_uint32 (buf, 0, LECO_MAGIC);
	lec_set_uint16 (buf, 4, 0);
	lec_set_uint16 (buf, 6, 80);
	lec_set_uint32 (buf, 8, lf->bsize);
	lec_set_uint32 (buf, 12, lf->wsize);
	lec_set_uint64 (buf, 16, lf->dofs);
	lec_set_uint64 (buf, 24, lf->eofs);
	lec_set_uint32 (buf, 32, lf->bcnt);
	lec_set_uint32 (buf, 36, lf->ecnt);
	lec_set_uint64 (buf, 40, lf->fsize);

	memcpy (buf + 48, lf->md5, 16);

	lec_set_uint32 (buf, 64, 0);
	lec_set_uint32 (buf, 68, lf->fcrc);
	lec_set_uint32 (buf, 72, lf->dcrc);
	lec_set_uint32 (buf, 76, 0);

	crc = lec_get_crc (buf, 80);
	lec_set_uint32 (buf, 76, crc);

	if (lecf_write (lf, 0, buf, LECO_HEADER) != LECO_HEADER) {
		return (1);
	}

	return (0);
}

int lecf_write_magic1 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC1, 8);
	lec_set_uint64 (buf, 8, lf->dofs - 16);

	if (lecf_write (lf, lf->dofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_magic2 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC2, 8);
	lec_set_uint64 (buf, 8, lf->eofs - 16);

	if (lecf_write (lf, lf->eofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_magic3 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC3, 8);
	lec_set_uint64 (buf, 8, lf->rofs - 16);

	if (lecf_write (lf, lf->rofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_magic4 (lec_file_t *lf)
{
	unsigned char buf[16];

	memcpy (buf, LECO_MAGIC4, 8);
	lec_set_uint64 (buf, 8, lf->zofs - 16);

	if (lecf_write (lf, lf->zofs - 16, buf, 16) != 16) {
		return (1);
	}

	return (0);
}

int lecf_write_redundant (lec_file_t *lf, const void *crc)
{
	unsigned      i;
	uint64_t      rofs;
	unsigned char buf[16];

	if (lecf_write_magic3 (lf)) {
		return (1);
	}

	if (lecf_write (lf, lf->rofs, &lf->header[0], LECO_HEADER) != LECO_HEADER) {
		return (1);
	}

	rofs = lf->rofs + LECO_HEADER;

	if (crc == NULL) {
		uint64_t dofs;

		dofs = lf->dofs;

		for (i = 0; i < lf->bcnt; i++) {
			if (lecf_read (lf, dofs, buf, 4) != 4) {
				return (1);
			}

			if (lecf_write (lf, rofs, buf, 4) != 4) {
				return (1);
			}

			dofs += 4;
			rofs += 4;
		}
	}
	else {
		unsigned long cnt;

		cnt = 4UL * lf->bcnt;
		if (lecf_write (lf, rofs, crc, cnt) != cnt) {
			return (1);
		}
	}

	lf->zofs = (lf->rofs + LECO_HEADER + 4UL * lf->bcnt + 15) & ~15UL;
	lf->zofs += 16;

	return (0);
}
