/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecoinfo.c                                             *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2007 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecoinfo.c 83 2007-08-05 18:27:40Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "lecofile.h"
#include "util.h"


const char *argv0 = NULL;

static int par_info_all = 1;
static int par_info_cmdline = 0;
static int par_info_dblk = 0;
static int par_info_eblk = 0;
static int par_info_first = 0;
static int par_info_hsize = 0;
static int par_info_md5 = 0;
static int par_info_ecccnt = 0;
static int par_info_redundant = 0;


static
void prt_version (void)
{
	fputs (
		"lecoinfo version " LECO_VERSION_STR
		" (" LECO_CFG_DATE " " LECO_CFG_TIME ")\n\n"
		"Copyright (C) 2005-2007 by Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_help (void)
{
	fputs (
		"Usage: lecoinfo [options] file\n"
		"  -a, --all               Print all information [default]\n"
		"  -c, --command-line      Print the command line [no]\n"
		"  -d, --data-blocks       List data blocks\n"
		"  -e, --ecc-blocks        List ECC blocks\n"
		"  -f, --first-block       Print first ECC block index\n"
		"  -h, --header-size       Print header size\n"
		"  -m, --md5               Print MD5 of data file\n"
		"  -n, --ecc-block-count   Print number of ECC blocks\n"
		"  -r, --redundant-header  Print redundant header status\n",
		stdout
	);

	fflush (stdout);
}

static
int lec_info_open (lec_file_t *lf, FILE *fp)
{
	if (lecf_open (lf, fp)) {
		return (1);
	}

	return (0);
}

static
int lec_info_check_redundant (lec_file_t *lf)
{
	unsigned      cnt;
	unsigned char buf[256];

	if (lecf_read (lf, lf->rofs - 16, buf, 16) != 16) {
		return (1);
	}

	if (memcmp (buf, LECO_MAGIC3, 8) != 0) {
		return (1);
	}

	cnt = (lf->hsize < 256) ? lf->hsize : 256;

	if (lecf_read (lf, lf->rofs, buf, cnt) != cnt) {
		return (1);
	}

	if (memcmp (lf->header, buf, cnt) != 0) {
		return (1);
	}

	return (0);
}

/*
 * Get the block number of the first ECC block in the file.
 */
static
int lec_info_get_first (lec_file_t *lf, unsigned long *val)
{
	unsigned char buf[4];

	if (lecf_read (lf, lf->eofs, buf, 4) != 4) {
		*val = 0;
		return (1);
	}

	*val = lec_get_uint32 (buf, 0);

	return (0);
}

static
int lec_info_all (lec_file_t *lf)
{
	unsigned   i;

	printf ("magic:            %08lX\n", lf->magic);
	printf ("format version:   %u\n", lf->version);
	printf ("header size:      %u\n", lf->hsize);
	printf ("file size:        %llu\n", lf->fsize);
	printf ("file crc:         %08lx (%08lx)\n", lf->fcrc, lf->dcrc);
	printf ("md5:              ");

	for (i = 0; i < 16; i++) {
		printf ("%02x", lf->md5[i]);
	}

	fputs ("\n", stdout);

	printf ("redundant header: %s\n", lec_info_check_redundant (lf) ? "no" : "ok");
	printf ("block size:       %lu words\n", lf->bsize);
	printf ("block size:       %lu bytes\n", lf->size);
	printf ("word size:        %u bits\n", lf->wsize);
	printf ("base:             0x%08llx\n", (unsigned long long) lf->base);
	printf ("header offset:    0x%08llx\n", (unsigned long long) lf->hofs);
	printf ("data crc offset:  0x%08llx\n", lf->dofs);
	printf ("ecc block offset: 0x%08llx\n", lf->eofs);
	printf ("data block count: %u\n", lf->bcnt);
	printf ("ecc block count:  %u\n", lf->ecnt);

	return (0);
}

/*
 * Print the command line used to create this ECC file.
 */
static
int lec_info_cmdline (lec_file_t *lf, const char *fname)
{
	unsigned long first;

	fputs ("leco -e", stdout);

	if (lec_info_check_redundant (lf) == 0) {
		fputs (" -r", stdout);
	}

	printf (" -n %u", lf->bcnt);
	printf (" -S %u", lf->ecnt);

	if (lec_info_get_first (lf, &first) == 0) {
		printf (" -f %lu", first);
	}

	printf (" -i source.dat -l %s\n", fname);

	return (0);
}

static
int lec_info_hsize (lec_file_t *lf)
{
	printf ("%u\n", lf->hsize);

	return (0);
}

static
int lec_info_ecccnt (lec_file_t *lf)
{
	printf ("%u\n", lf->ecnt);

	return (0);
}

static
int lec_info_md5 (lec_file_t *lf)
{
	unsigned i;

	for (i = 0; i < 16; i++) {
		printf ("%02x", lf->md5[i]);
	}

	fputs ("\n", stdout);

	return (0);
}

static
int lec_info_eblk (lec_file_t *lf)
{
	unsigned long i;
	unsigned char buf[8];
	unsigned long blk, crc;
	uint64_t      ofs;

	ofs = lf->eofs;

	for (i = 0; i < lf->ecnt; i++) {
		if (lecf_read (lf, ofs, buf, 4) != 4) {
			return (1);
		}
		if (lecf_read (lf, ofs + lf->size + 4, buf + 4, 4) != 4) {
			return (1);
		}

		blk = lec_get_uint32 (buf, 0);
		crc = lec_get_uint32 (buf, 4);

		printf ("%lu\t%lu\t%08lx\n", i, blk, crc);

		ofs += lf->size + 8;
	}

	return (0);
}

static
int lec_info_dblk (lec_file_t *lf)
{
	unsigned long i;
	unsigned char buf[4];
	unsigned long crc;
	uint64_t      ofs;

	ofs = lf->dofs;

	for (i = 0; i < lf->bcnt; i++) {
		if (lecf_read (lf, ofs, buf, 4) != 4) {
			return (1);
		}

		crc = lec_get_uint32 (buf, 0);

		printf ("%lu\t%08lx\n", i, crc);

		ofs += 4;
	}

	return (0);
}

static
int lec_info_first (lec_file_t *lf)
{
	unsigned long blk;

	if (lec_info_get_first (lf, &blk)) {
		return (1);
	}

	printf ("%lu\n", blk);

	return (0);
}

static
int lec_info_redundant (lec_file_t *lf)
{
	if (lec_info_check_redundant (lf)) {
		fputs ("0\n", stdout);
	}
	else {
		fputs ("1\n", stdout);
	}

	return (0);
}

static
int lec_info_fp (FILE *fp, const char *fname)
{
	lec_file_t lf;

	if (lec_info_open (&lf, fp)) {
		return (1);
	}

	if (par_info_all) {
		lec_info_all (&lf);
	}
	else if (par_info_cmdline) {
		lec_info_cmdline (&lf, fname);
	}
	else if (par_info_dblk) {
		lec_info_dblk (&lf);
	}
	else if (par_info_eblk) {
		lec_info_eblk (&lf);
	}
	else if (par_info_first) {
		lec_info_first (&lf);
	}
	else if (par_info_hsize) {
		lec_info_hsize (&lf);
	}
	else if (par_info_ecccnt) {
		lec_info_ecccnt (&lf);
	}
	else if (par_info_md5) {
		lec_info_md5 (&lf);
	}
	else if (par_info_redundant) {
		lec_info_redundant (&lf);
	}

	return (0);
}

static
int lec_info (const char *lec)
{
	int  r;
	FILE *fp;

	fp = fopen (lec, "rb");
	if (fp == NULL) {
		return (1);
	}

	r = lec_info_fp (fp, lec);

	fclose (fp);

	return (r);
}

int main (int argc, char **argv)
{
	int        i, r;
	const char *fname;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg (argv[1], NULL, "--version")) {
			prt_version();
			return (0);
		}
		else if (str_isarg (argv[1], NULL, "--help")) {
			prt_help();
			return (0);
		}
	}

	fname = NULL;

	i = 1;
	while (i < argc) {
		if (str_isarg (argv[i], "-a", "--all")) {
			par_info_all = 1;
		}
		else if (str_isarg (argv[i], "-c", "--command-line")) {
			par_info_all = 0;
			par_info_cmdline = 1;
		}
		else if (str_isarg (argv[i], "-d", "--data-blocks")) {
			par_info_all = 0;
			par_info_dblk = 1;
		}
		else if (str_isarg (argv[i], "-e", "--ecc-blocks")) {
			par_info_all = 0;
			par_info_eblk = 1;
		}
		else if (str_isarg (argv[i], "-f", "--first-block")) {
			par_info_all = 0;
			par_info_first = 1;
		}
		else if (str_isarg (argv[i], "-h", "--header-size")) {
			par_info_all = 0;
			par_info_hsize = 1;
		}
		else if (str_isarg (argv[i], "-m", "--md5")) {
			par_info_all = 0;
			par_info_md5 = 1;
		}
		else if (str_isarg (argv[i], "-n", "--ecc-block-count")) {
			par_info_all = 0;
			par_info_ecccnt = 1;
		}
		else if (str_isarg (argv[i], "-r", "--redundant-header")) {
			par_info_all = 0;
			par_info_redundant = 1;
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv[0], argv[i]
			);
			return (1);
		}
		else {
			if (fname == NULL) {
				fname = argv[i];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], argv[i]
				);
				return (1);
			}
		}

		i += 1;
	}

	if (fname != NULL) {
		r = lec_info (fname);
	}
	else {
		r = 1;
	}

	return (r);
}
